/**
 * JASMINe
 * Copyright (C) 2009 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: NotificationService.java 4836 2009-08-09 20:41:41Z renaultgu $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.notification.service;

import java.io.File;
import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.ow2.jasmine.event.beans.JasmineEventNotification;
import org.ow2.jasmine.event.beans.JasmineEventNotificationExecutedAction;
import org.ow2.jasmine.monitoring.eos.common.jms.JMSConnection;
import org.ow2.jasmine.monitoring.eos.common.jndilocator.ServiceLocator;
import org.ow2.jasmine.monitoring.eos.notification.actions.api.IAction;
import org.ow2.jasmine.monitoring.eos.notification.actions.exchange.ActionResult;
import org.ow2.jasmine.monitoring.eos.notification.filters.api.IFilter;
import org.ow2.jasmine.monitoring.eos.notification.service.exchange.AcknowledgeState;
import org.ow2.jasmine.monitoring.eos.notification.service.exchange.ActionCall;
import org.ow2.jasmine.monitoring.eos.notification.service.exchange.ExecutedAction;
import org.ow2.jasmine.monitoring.eos.notification.service.exchange.LevelSetting;
import org.ow2.jasmine.monitoring.eos.notification.service.exchange.NotificationSetting;
import org.ow2.jasmine.monitoring.eos.notification.service.jms.JMSMessageListener;
import org.ow2.jasmine.monitoring.eos.notification.service.jms.NotificationEventManager;
import org.ow2.jasmine.monitoring.eventswitch.beans.JasmineEventNotificationSLBRemote;
import org.ow2.jonas.lib.security.context.SecurityCurrent;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * @author radua
 */
public class NotificationService implements Serializable {
    /**
     * Service class for the notification module. It connects to the database
     * and retrieves all the notifications.
     */

    // ////////////////////////////////////////
    // General variables
    // ///////////////////////////////////////
    /**
     * serialVersionUID.
     */
    private static final long serialVersionUID = 1L;

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(NotificationService.class);

    /**
     * Notification configuration file path.
     */
    private static final String NOTIFICATION_CONFIG = "META-INF/notification-config.xml";

    /**
     * Notification configuration xsd file path.
     */
    private static final String NOTIFICATION_CONFIG_XSD = "resources/notification-config.xsd";

    /**
     * Default Acknowledge class to use.
     */
    private static final String DEFAULT_ACKNOWLEDGE_ACTION_CLASS = "org.ow2.jasmine.monitoring.eos.notification.actions.impl.AcknowledgeAction";

    /**
     * Name of the acknowledge action name.
     */
    private String currentAcknowledgeActionName;

    /**
     * HashMap of the actions. The key is the name, the object the className.
     */
    private Map<String, String> availableActions = new HashMap<String, String>();

    /**
     * The list of available levels. The key is the level name.
     */
    private Map<String, LevelSetting> availableLevels = new HashMap<String, LevelSetting>();

    /**
     * The map that contains all the class names of the defined filters. The key
     * is the class name, the value is whether or not the filter must be
     * processed when reviewing datas.
     */
    private Map<String, Boolean> filters = new HashMap<String, Boolean>();

    private List<IFilter> alwaysProcessedFilters = new ArrayList<IFilter>();

    private List<IFilter> onlyWhenLiveProcessedFilters = new ArrayList<IFilter>();

    // ////////////////////////////////////////
    // Variables concerning database
    // ///////////////////////////////////////
    /**
     * Get remote instance of the ejb interface through the ServiceLocator.
     */

    private static JasmineEventNotificationSLBRemote notifFacade = (JasmineEventNotificationSLBRemote) ServiceLocator
        .getInstance().getRemoteObject("db-ejb/notification");

    // ////////////////////////////////////////
    // Variables concerning JMS
    // ///////////////////////////////////////
    /**
     * The JMS connection instance.
     */
    private JMSConnection connection = null;

    /**
     * The event manager instance.
     */
    private NotificationEventManager manager = null;

    // ////////////////////////////////////////
    // General methods
    // ///////////////////////////////////////
    public NotificationService() {
        if (this.manager == null) {
            this.manager = NotificationEventManager.getInstance();
        }
    }

    /**
     * Get all the notifications from the database.
     *
     * @return a List of JasmineEventNotification objects.
     */

    // ////////////////////////////////////////
    // Methods concerning DB
    // ///////////////////////////////////////
    /**
     * Gets all the notification from the database. Only done when in live mode.
     */
    public List<JasmineEventNotification> getNotifsFromDB() {

        // Disconnect from the topic if connected. it will be reconnected
        // automatically later, when first getting notifications from the topic
        this.disconnectTopic();

        logger.debug("Getting notification from Data Base");

        JasmineEventNotification[] notifList = null;
        List<JasmineEventNotification> result = new ArrayList<JasmineEventNotification>();

        notifList = notifFacade.getEventNotifications(null, null, null, null, null, null, null, null);

        if (notifList != null) {
            for (JasmineEventNotification j : notifList) {
                // if the notification must be send, then add it to the result
                // list
                if (filter(j, false)) {
                    result.add(j);
                }
            }
        }

        return result;
    }

    public List<JasmineEventNotification> getNotifsFromDBFromTo(final Date dateFrom, final Date dateTo) {

        // Disconnect from the topic if connected. We're in a replay mode.
        this.disconnectTopic();

        logger.debug("Getting notification from Data Base from date {0} to date {1}.", dateFrom, dateTo);

        List<JasmineEventNotification> notifList = null;
        notifList = notifFacade.getEventNotifications(dateFrom, dateTo);

        List<JasmineEventNotification> result = new ArrayList<JasmineEventNotification>();

        if (notifList != null) {
            for (JasmineEventNotification j : notifList) {
                // if the notification must be send, then add it to the result
                // list
                if (filter(j, true)) {
                    result.add(j);
                }
            }
        }

        notifList = null;

        return result;
    }

    // ////////////////////////////////////////
    // Methods concerning JMS
    // ///////////////////////////////////////
    /**
     * Get all the notifications from the JMS topic.
     *
     * @return a List of JasmineEventNotification objects.
     */

    public List<JasmineEventNotification> getNotifsFromTopic() {

        logger.debug("Getting notification from JMS Topic");

        if (this.connection == null) {
            connectTopic();
        }
        List<JasmineEventNotification> list = new ArrayList<JasmineEventNotification>();
        list.addAll(this.manager.getAll());

        logger.debug("Received {0} notifications.", list.size());

        List<JasmineEventNotification> result = new ArrayList<JasmineEventNotification>();

        for (JasmineEventNotification j : list) {
            // if the notification must be send, then add it to the result list
            if (filter(j, false)) {
                result.add(j);
            }
        }

        list = null;

        return result;
    }

    /**
     * Connect to the JMS topic.
     *
     * @return -1 if an error occured, 0 otherwise.
     */
    public int connectTopic() {
        if (this.connection == null) {
            this.connection = new JMSConnection(this.manager);
        }

        logger.debug("Connected to the topic ...");

        return this.connection.connect("jasmineNotification", new JMSMessageListener(this.manager));
    }

    /**
     * Disconnect from the JMS topic.
     *
     * @return -1 if an error occurred, 0 otherwise.
     */
    public int disconnectTopic() {
        logger.debug("Disconnecting from the JMS topic...");
        try {

            if (this.connection == null) {
                return -1;
            }
            return this.connection.disconnect();
        } catch (Exception e) {
            return -1;
        }
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // Common methods for both JMS and DB
    //
    // ////////////////////////////////////////////////////////////////////////

    private boolean filter(final JasmineEventNotification notification, final boolean reviewing) {

        boolean result = true;

        for (IFilter filter : this.alwaysProcessedFilters) {
            result &= filter.filter(notification);
        }

        // if not reviewing, execute the live filters
        if (!reviewing) {
            for (IFilter filter : this.onlyWhenLiveProcessedFilters) {
                result &= filter.filter(notification);
            }
        }

        return result;
    }

    /**
     * Initialize the filter classes.
     *
     * @throws Exception
     */
    private void initializeFilters() throws Exception {

        logger.info("Initialize filters");

        Set<String> filters = this.filters.keySet();

        for (String className : filters) {
            Class<?> filterClass = Class.forName(className);
            Constructor<?> filterConstructor = filterClass.getConstructor();

            IFilter filter = (IFilter) filterConstructor.newInstance();

            if (this.filters.get(className).booleanValue()) {
                logger.info("Add {0} to the alwaysExecutedFilter list", className);
                this.alwaysProcessedFilters.add(filter);
            } else {
                logger.info("Add {0} to the onlyWhenLiveProcessedFilters list", className);
                this.onlyWhenLiveProcessedFilters.add(filter);
            }
        }
    }

    // ////////////////////////////////////////
    // Actions
    // ///////////////////////////////////////
    public List<NotificationSetting> getActions() {

        List<NotificationSetting> actionsList = new ArrayList<NotificationSetting>();

        try {

            URL notification = this.getClass().getClassLoader().getResource(NOTIFICATION_CONFIG);

            logger.debug("loading file : {0}", notification);

            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document doc = null;

            // validate the xml file
            // create a SchemaFactory capable of understanding WXS schemas
            SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

            // load a WXS schema, represented by a Schema instance
            Source schemaFile = new StreamSource(new File(this.getClass().getClassLoader().getResource(NOTIFICATION_CONFIG_XSD)
                .toURI()));
            Schema schema = factory.newSchema(schemaFile);

            // create a Validator instance, which can be used to validate an
            // instance document
            Validator validator = schema.newValidator();

            // validate the DOM tree
            try {
                validator.validate(new DOMSource(doc));
            } catch (SAXException e) {
                // instance document is invalid!
                throw new Exception("The configuration file " + NOTIFICATION_CONFIG
                    + " is not well formed. Please refer to the XML Schema.", e);
            }

            ClassLoader cl = this.getClass().getClassLoader();
            if (cl.getResource(NOTIFICATION_CONFIG) != null) {

                logger.info("loading file : {0}", NOTIFICATION_CONFIG);

                logger.debug("Loading the notification XML file from classpath");
                InputStream resource = cl.getResourceAsStream(NOTIFICATION_CONFIG);
                try {
                    doc = documentBuilder.parse(resource);
                } finally {
                    resource.close();
                    resource = null;
                }
            } else {
                logger.error("No configuration file available : {0}", NOTIFICATION_CONFIG);
            }

            if (doc != null) {
                doc.getDocumentElement().normalize();

                // Analyse the filters
                NodeList listOfFiltersEntries = doc.getElementsByTagName("filters");
                Node filters = listOfFiltersEntries.item(0);

                // <filter> nodes
                NodeList filterList = filters.getChildNodes();
                for (int filterListLength = 0; filterListLength < filterList.getLength(); filterListLength++) {
                    Node filter = filterList.item(filterListLength);
                    if ("filter".equals(filter.getNodeName())) {
                        String className = filter.getAttributes().getNamedItem("class").getTextContent();
                        Boolean reviewProcessed = new Boolean(filter.getAttributes().getNamedItem("reviewProcess")
                            .getTextContent());
                        this.filters.put(className, reviewProcessed);
                    }
                }

                // Acknowledge Class configuration
                NodeList listOfAckClasseEntries = doc.getElementsByTagName("acknowledgeClass");
                Node ackClasses = listOfAckClasseEntries.item(0);

                // <acknowledgeClass> node
                String ackClassSet = ackClasses.getAttributes().getNamedItem("class").getTextContent();
                this.currentAcknowledgeActionName = ackClasses.getAttributes().getNamedItem("name").getTextContent();
                if ("default".equals(ackClassSet)) {
                    this.availableActions.put(this.currentAcknowledgeActionName, DEFAULT_ACKNOWLEDGE_ACTION_CLASS);
                } else {
                    this.availableActions.put(this.currentAcknowledgeActionName, ackClassSet);
                }

                // keepDisplayed configuration
                NodeList listOfkeepDisplayedEntries = doc.getElementsByTagName("keepDisplayed");
                Node keepDisplayed = listOfkeepDisplayedEntries.item(0);

                // <keepDisplayed> node
                String numberOfHours = keepDisplayed.getAttributes().getNamedItem("hours").getTextContent();

                // Analyse the levels
                NodeList listOfLevelsEntries = doc.getElementsByTagName("levels");
                Node levels = listOfLevelsEntries.item(0);

                // <level> nodes
                NodeList levelList = levels.getChildNodes();
                for (int levelListLength = 0; levelListLength < levelList.getLength(); levelListLength++) {
                    LevelSetting levelSetting = new LevelSetting();
                    Node level = levelList.item(levelListLength);
                    if ("level".equals(level.getNodeName())) {
                        String name = level.getAttributes().getNamedItem("name").getTextContent();
                        String color = level.getAttributes().getNamedItem("color").getTextContent();
                        levelSetting.setLevel(name);
                        if (!"".equals(color)) {
                            levelSetting.setColor(color);
                        }
                        logger.debug("name : ", level.getAttributes().getNamedItem("name").getTextContent());
                        logger.debug("color : ", level.getAttributes().getNamedItem("color").getTextContent());
                    }
                    this.availableLevels.put(levelSetting.getLevel(), levelSetting);
                }

                // Retrieve the <notifs> nodes
                NodeList listOfEntries = doc.getElementsByTagName("notifs");
                for (int j = 0; j < listOfEntries.getLength(); j++) {
                    Node notifs = listOfEntries.item(j);
                    logger.debug("<notifs> node name : {0}", notifs.getNodeName());

                    NodeList allNotifs = notifs.getChildNodes();

                    // for all <notif> nodes
                    for (int k = 0; k < allNotifs.getLength(); k++) {
                        Node notif = allNotifs.item(k);
                        logger.debug("<notif> node name : {0}", notif.getNodeName());
                        if ("All".equals(notif.getNodeName())) {
                            logger.info("Getting all the notifications ... No levels will be be associated then.");

                            // returning here will aim at showing all the
                            // notifications (the list is empty here)
                            return actionsList;

                        }
                        if ("notif".equals(notif.getNodeName())) {

                            NotificationSetting actionBinding = new NotificationSetting();
                            // Set the rule id of the actionBinding object.
                            actionBinding.setRule(notif.getAttributes().getNamedItem("rule").getTextContent());

                            // Set the number of hours that the notification
                            // should stay in the board.
                            actionBinding.setKeepDisplayed(Integer.valueOf(numberOfHours));

                            // set the level associated to the rule
                            if (this.availableLevels.get(notif.getAttributes().getNamedItem("level").getTextContent()) != null) {
                                actionBinding.setLevel(this.availableLevels.get(notif.getAttributes().getNamedItem("level")
                                    .getTextContent()));
                            } else {
                                throw new Exception("Level " + notif.getAttributes().getNamedItem("level").getTextContent()
                                    + " not defined in the <levels> tag");
                            }

                            // set if the acknowledge is required. If so, set
                            // the name of the acknowledge button
                            boolean synch = Boolean
                                .valueOf(notif.getAttributes().getNamedItem("synchronized").getTextContent()).booleanValue();
                            actionBinding.setSynchronize(synch);

                            actionBinding.setAcknowledgeActionName(this.currentAcknowledgeActionName);

                            // retrieve the <actions> node
                            NodeList allActions = notif.getChildNodes();
                            for (int m = 0; m < allActions.getLength(); m++) {
                                Node actions = allActions.item(m);
                                logger.debug("<actions> node name : {0}", actions.getNodeName());
                                if ("actions".equals(actions.getNodeName())) {
                                    NodeList allAction = actions.getChildNodes();

                                    // for all the <action> nodes
                                    for (int l = 0; l < allAction.getLength(); l++) {
                                        Node action = allAction.item(l);
                                        logger.debug("<action> node name : {0}", action.getNodeName());
                                        if ("action".equals(action.getNodeName())) {

                                            actionBinding.addAction(action.getAttributes().getNamedItem("name")
                                                .getTextContent());

                                            if (!availableActions.containsKey(action.getAttributes().getNamedItem("name")
                                                .getTextContent())) {
                                                availableActions.put(action.getAttributes().getNamedItem("name")
                                                    .getTextContent(), action.getAttributes().getNamedItem("className")
                                                    .getTextContent());
                                            }
                                        }

                                    }// End for all the <action> nodes

                                }
                            } // End retrieve the <actions> node

                            logger.debug("number of actions in the binding object : {0}", actionBinding.getAction().size());

                            actionsList.add(actionBinding);

                        }
                    }
                }

            }

            // initialize filters
            this.initializeFilters();

        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();

        }

        return actionsList;
    }

    /**
     * Execute the required action.
     *
     * @param actionName the action name to execute.
     * @param jasmineEventNotification the notification associated to the
     *        action.
     * @return the executed action name.
     * @throws Exception All exceptions are thrown.
     */
    public ExecutedAction execute(final ActionCall action) throws Exception {

        try {

            if (action.getName().equals(this.currentAcknowledgeActionName)) {
                logger.debug("Update notification {0} to the acknowledged state", action.getNotification().getId());
                notifFacade.updateNotificationStatus(action.getNotification().getId(), true);
            }

            JasmineEventNotificationExecutedAction actionToSave = new JasmineEventNotificationExecutedAction(action
                .getComment(), action.getName(), new Date(), action.getUser());

            notifFacade.saveExecutedAction(action.getNotification().getId(), actionToSave);

            logger.debug("Action name : {0} , id : {1} .\n\tThe associated class is : {2}", action.getName(), action
                .getNotification().getId(), this.availableActions.get(action.getName()));

            Class<?> actionClass = Class.forName(this.availableActions.get(action.getName()));

            Constructor<?> actionConstructor = actionClass.getConstructor();
            IAction actionInstance = (IAction) actionConstructor.newInstance();

            logger.info("Executing the following action : {0}", actionInstance.getName());

            ActionResult actionResult = actionInstance.execute(action.getNotification());
            actionResult.setName(action.getName());

            ExecutedAction executedAction = new ExecutedAction();
            executedAction.setActionResult(actionResult);
            executedAction.setUser(action.getUser());
            executedAction.setNotification(notifFacade.getEventNotification(action.getNotification().getId()));

            return executedAction;
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw e;
        }

    }

    /**
     * Check is the notification has been acknowledge.
     *
     * @param notification the notification to check.
     * @return the acknowledge state.
     */
    public AcknowledgeState isAcknowledged(final JasmineEventNotification notification) {

        SecurityCurrent secutiryCurrent = SecurityCurrent.getCurrent();

        logger.debug("logged user : {0}", secutiryCurrent.getSecurityContext().getPrincipalName());

        logger.debug("Checking notification {0} from rule {1}", notification.getId(), notification.getRule());

        JasmineEventNotification notif = notifFacade.getEventNotification(notification.getId());

        AcknowledgeState ackState = null;

        if (notif.getStatus()) {
            for (JasmineEventNotificationExecutedAction action : notif.getActions()) {
                if (action.getAction().equals(this.currentAcknowledgeActionName)) {
                    ackState = new AcknowledgeState(notif.getStatus(), action.getUser(), notif.getId());
                }
            }
        } else {
            ackState = new AcknowledgeState(notif.getStatus(), "", notif.getId());
        }

        ackState.setLoggedUser(secutiryCurrent.getSecurityContext().getPrincipalName());

        logger.debug("Notification {0} checked with status {1}.", ackState.getNotificationId(), ackState.getState());

        return ackState;
    }

    /**
     * Get the logged user.
     *
     * @return the logged user.
     */
    public String getLoggedUser() {
        return SecurityCurrent.getCurrent().getSecurityContext().getPrincipalName();
    }
}
